@@ -1,29 +1,22 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-# (c) Copyright IBM Corp. 2010, 2019. All Rights Reserved.
+# (c) Copyright IBM Corp. 2010, 2023. All Rights Reserved.
 
 """Simple client for Resilient REST API"""
-from __future__ import print_function
-
+import datetime
+import importlib
 import json
-import ssl
-import mimetypes
+import logging
 import os
 import sys
-import logging
-import datetime
-import unicodedata
-import requests
-import importlib
-from . import co3base
-from .patch import PatchStatus
 from argparse import Namespace
-from requests.adapters import HTTPAdapter
-from requests.packages.urllib3.poolmanager import PoolManager
-from requests_toolbelt.multipart.encoder import MultipartEncoder
-from cachetools import cachedmethod
-from cachetools.ttl import TTLCache
-from .co3base import ensure_unicode, get_proxy_dict, NoChange
+
+import requests
+from cachetools import TTLCache, cachedmethod
+
+from . import co3base
+from .co3base import NoChange, ensure_unicode, get_proxy_dict
+from .patch import PatchStatus
 
 try:
     # Python 3
@@ -97,13 +90,19 @@ def get_config_file(filename=None, generate_filename=False):
     return config_file
 
 
-def get_client(opts):
+def get_client(opts, custom_headers=None, **kwargs):
     """
-    Helper: get a SimpleClient for Resilient REST API.
+    This is a helper method to get an instance of
+    :class:`SimpleClient` for the SOAR REST API.
 
-    :param opts: the connection options, as a :class:`dict`, or a :class:`Namespace`
+    :param opts: the connection options - usually the contents of the app.config file
+    :type opts: dict
+    :param custom_headers: A dictionary of any headers you want to send in **every** request
+    :type custom_headers: dict
+    :param kwargs: A dictionary of any other keyword arguments
+    :type kwargs: dict
 
-    A standard way to initialize a SimpleClient with default configuration is,
+    A standard way to initialize a :class:`SimpleClient` with default configuration is:
 
     .. code-block:: python
 
@@ -111,7 +110,7 @@ def get_client(opts):
         opts = parser.parse_args()
         client = resilient.get_client(opts)
 
-    Returns: a connected and verified instance of SimpleClient.
+    :return: a connected and verified instance of :class:`SimpleClient`
     """
     if isinstance(opts, Namespace):
         opts = vars(opts)
@@ -123,6 +122,13 @@ def get_client(opts):
         requests.packages.urllib3.disable_warnings()  # otherwise things get very noisy
         verify = False
 
+    # Enable client certificate authentication
+    clientauthcert = opts.get("client_auth_cert", None)
+    clientauthkey = opts.get("client_auth_key", None)
+    certauth = (clientauthcert, clientauthkey)
+    if str(clientauthcert).lower() == "false" or clientauthcert is None or clientauthkey is None:
+        certauth = False
+
     proxy = None
     if opts.get("proxy_host"):
         proxy = get_proxy_dict(opts)
@@ -133,7 +139,9 @@ def get_client(opts):
     simple_client_args = {"org_name": opts.get("org"),
                           "proxies": proxy,
                           "base_url": url,
-                          "verify": verify}
+                          "verify": verify,
+                          "certauth": certauth,
+                          "custom_headers": custom_headers}
     if opts.get("log_http_responses"):
         LOG.warning("Logging all HTTP Responses from Resilient to %s", opts["log_http_responses"])
         simple_client = LoggingSimpleClient
@@ -141,6 +149,9 @@ def get_client(opts):
     else:
         simple_client = SimpleClient
 
+    # Update with kwargs
+    simple_client_args.update(kwargs)
+
     resilient_client = simple_client(**simple_client_args)
 
     if opts.get("resilient_mock"):
@@ -213,28 +224,119 @@ def _raise_if_error(response):
 
 
 class SimpleClient(co3base.BaseClient):
-    """Python helper class for using the Resilient REST API."""
+    """
+    Python helper class for using the IBM SOAR REST API.
 
-    def __init__(self, org_name=None, base_url=None, proxies=None, verify=None, cache_ttl=240):
+    .. code-block::
+
+        '''
+        An example script showing how to connect and get
+        incidents with IBM Security SOAR
+
+        Example Usage:
+        python sample_connect.py "https://<host>", "<org_name>", "<api_key_id>", "<api_key_secret>", "<path_to_ca_file>|False"
+        '''
+
+        import sys
+        import os
+        from resilient import SimpleClient
+
+
+        def main():
+
+            args = sys.argv
+            assert len(args) == 6
+            SCRIPT_NAME = args[0]
+
+            # Get required parameters
+            HOST = args[1]
+            ORG_NAME = args[2]
+            API_KEY_ID = args[3]
+            API_KEY_SECRET = args[4]
+
+            # Check if using cafile
+            ca_file = args[5]
+            VERIFY = ca_file if os.path.isfile(ca_file) is True else False
+
+            # Instansiate SimpleClient
+            res_client = SimpleClient(
+                org_name=ORG_NAME,
+                base_url=HOST,
+                verify=VERIFY
+            )
+
+            # Set the API Key
+            res_client.set_api_key(
+                api_key_id=API_KEY_ID,
+                api_key_secret=API_KEY_SECRET
+            )
+
+            # Setup the request payload
+            payload = {
+                "filters": [
+                    {
+                        "conditions": [{"field_name": "plan_status", "method": "equals", "value": "A"}]
+                    }
+                ]
+            }
+
+            # Invoke the request
+            response = res_client.post("/incidents/query_paged?return_level=full", payload=payload)
+
+            # Print results
+            for incident in response.get("data", []):
+                print("{0}: {1}".format(incident.get("id"), incident.get("name")))
+
+
+        if __name__ == "__main__":
+            main()
+
+
+    .. note::
+        The full REST API Documentation for IBM SOAR can be
+        viewed on the Platform itself by searching for the **REST API Reference**
+        or by going to: ``https://<base_url>/docs/rest-api/index.html``
+    """
+
+    def __init__(self, org_name=None, base_url=None, proxies=None, verify=None, cache_ttl=240, certauth=None, custom_headers=None, **kwargs):
         """
-
         :param org_name: The name of the organization to use.
-        :param base_url: The base URL of the Resilient server, e.g. 'https://app.resilientsystems.com/'
-        :param proxies: A dictionary of HTTP proxies to use, if any.
-        :param verify: The path to a PEM file containing the trusted CAs, or False to disable all TLS verification
-        :param cache_ttl: Time to live for cached API responses
+        :type org_name: str
+        :param base_url: The base URL of the SOAR server, e.g. ``https://soar.ibm.com/``
+        :type base_url: str
+        :param proxies: A dictionary of ``HTTP`` proxies to use, if any.
+        :type proxies: dict
+        :param verify: The path to a ``PEM`` file containing the trusted CAs, or ``False`` to disable all TLS verification
+        :type verify: str|bool
+        :param cache_ttl: Time in seconds to live for cached API responses
+        :type cache_ttl: int
+        :param certauth: The filepath for the client side certificate and the private key either as a single file or as a tuple of both files' paths
+        :type certauth: str|tuple(str, str)
+        :param custom_headers: A dictionary of any headers you want to send in **every** request
+        :type custom_headers: dict
+        :param kwargs: A dictionary of any other keyword arguments including;
+
+            * **max_connection_retries** (*int*) - Number of attempts to retry when connecting to SOAR. Use ``-1`` for unlimited retries. Defaults to ``-1``.
+            * **request_max_retries** (*int*) - Max number of times to retry a request to SOAR before exiting. Defaults to ``5``.
+            * **request_retry_delay** (*int*) - Number of seconds to wait between retries. Defaults to ``2``.
+            * **request_retry_backoff** (*int*) - Multiplier applied to delay between retry attempts. Defaults to ``2``.
+
+        :type kwargs: dict
         """
-        super(SimpleClient, self).__init__(org_name, base_url, proxies, verify)
+        super(SimpleClient, self).__init__(org_name, base_url, proxies, verify, certauth, custom_headers=custom_headers, **kwargs)
         self.cache = TTLCache(maxsize=128, ttl=cache_ttl)
 
     def connect(self, email, password, timeout=None):
         """
-        Connect and authenticate to the Resilient REST API service.
+        Connect and authenticate to the IBM SOAR REST API service.
 
         :param email: The email address to use for authentication.
+        :type email: str
         :param password: The password.
+        :type password: str
         :param timeout: optional timeout (seconds)
-        :return: The Resilient session object.
+        :type timeout: int
+        :return: The IBM SOAR session object.
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         # Calling connect of BaseClient. Convert the exception if there is any
@@ -256,95 +358,120 @@ class SimpleClient(co3base.BaseClient):
     def _get_cache(self):
         return self.cache
 
-    def get(self, uri, co3_context_token=None, timeout=None):
+    def get(self, uri, co3_context_token=None, timeout=None, is_uri_absolute=None, get_response_object=None, skip_retry=[]):
         """Gets the specified URI.
 
-        Note that this URI is relative to :samp:`<base_url>/rest/orgs/<org_id>`.  So for example,
-        if you specify a uri of :samp:`/incidents`, the actual URL would be something like:
-        `https://app.resilientsystems.com/rest/orgs/201/incidents`
+        .. note::
+            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
+            if you specify a uri of ``/incidents``, the actual URL would be something like:
+            ``https://soar.ibm.com/rest/orgs/201/incidents``
 
         :param uri: Relative URI of the resource to fetch.
-        :param co3_context_token: The Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :return: A dictionary or array with the value returned by the server.
+        :type uri: str
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :param is_uri_absolute: if True, does not insert /org/{org_id} into the uri.
+        :type is_uri_absolute: bool
+        :param get_response_object: if True, returns entire response object.
+        :type get_response_object: bool
+        :return: A dictionary, list, or response object with the value returned by the server.
+        :rtype: dict | list | Response
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         # Call get from BaseClient, convert exception if there is any
         response = None
         try:
-            response = super(SimpleClient, self).get(uri, co3_context_token, timeout)
+            response = super(SimpleClient, self).get(uri, co3_context_token, timeout, is_uri_absolute, get_response_object, skip_retry=skip_retry)
         except co3base.BasicHTTPException as ex:
             _raise_if_error(ex.get_response())
         return response
 
     @cachedmethod(_get_cache, key=_keyfunc)
-    def cached_get(self, uri, co3_context_token=None, timeout=None):
+    def cached_get(self, uri, co3_context_token=None, timeout=None, skip_retry=[]):
         """ Same as :meth:`get()`, but checks cache first """
-        return self.get(uri, co3_context_token, timeout)
+        return self.get(uri, co3_context_token, timeout, skip_retry=skip_retry)
 
     def get_const(self, co3_context_token=None, timeout=None):
         """
-        Get the ConstREST endpoint.
-        Endpoint for retrieving various constant information for this server.   This information is
-        useful in translating names that the user sees to IDs that other REST API endpoints accept.
-        For example, the incidentDTO has a field called "crimestatus_id". The valid values are stored
-        in constDTO.crime_statuses.
+        Get the ``ConstREST`` endpoint.
 
-        :param co3_context_token: The Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :return: ConstDTO as a dictionary
+        Endpoint for retrieving various constant information for this server. This information is
+        useful in translating names that the user sees to IDs that other REST API endpoints accept.
+
+        For example, the ``incidentDTO`` has a field called ``"crimestatus_id"``. The valid values are stored
+        in ``constDTO.crime_statuses``.
+
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :return: ``ConstDTO`` as a dictionary
+        :rtype: dict
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
-        url = u"{0}/rest/const".format(self.base_url)
-        response = self._execute_request(self.session.get,
-                                         url,
-                                         proxies=self.proxies,
-                                         cookies=self.cookies,
-                                         headers=self.make_headers(co3_context_token),
-                                         verify=self.verify,
-                                         timeout=timeout)
-        _raise_if_error(response)
-        return json.loads(response.text)
+        # Call get_const from BaseClient. Convert exception if there is any
+        response = None
+        try:
+            response = super(SimpleClient, self).get_const(co3_context_token, timeout)
+        except co3base.BasicHTTPException as ex:
+            _raise_if_error(ex.get_response())
+        return response
 
-    def get_content(self, uri, co3_context_token=None, timeout=None):
+    def get_content(self, uri, co3_context_token=None, timeout=None, skip_retry=[]):
         """Gets the specified URI.
 
-        Note that this URI is relative to :samp:`<base_url>/rest/orgs/<org_id>`.  So for example,
-        if you specify a uri of :samp:`/incidents`, the actual URL would be something like:
-        `https://app.resilientsystems.com/rest/orgs/201/incidents`
+        .. note::
+            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
+            if you specify a uri of ``/incidents``, the actual URL would be something like:
+            ``https://soar.ibm.com/rest/orgs/201/incidents``
 
         :param uri: Relative URI of the resource to fetch.
-        :param co3_context_token: The Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
+        :type uri: str
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
         :return: The raw value returned by the server for this resource.
+        :rtype: dict
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         # Call get_content from BaseClient. Convert exception if there is any
         response = None
         try:
-            response = super(SimpleClient, self).get_content(uri, co3_context_token, timeout)
+            response = super(SimpleClient, self).get_content(uri, co3_context_token, timeout, skip_retry=skip_retry)
         except co3base.BasicHTTPException as ex:
             _raise_if_error(ex.get_response())
         return response
 
-    def post(self, uri, payload, co3_context_token=None, timeout=None):
-        """Posts to the specified URI.
+    def post(self, uri, payload, co3_context_token=None, timeout=None, headers=None, skip_retry=[], **kwargs):
+        """
+        Posts to the specified URI.
 
-        Note that this URI is relative to :samp:`<base_url>/rest/orgs/<org_id>`.  So for example,
-        if you specify a uri of :samp:`/incidents`, the actual URL would be something like:
-        `https://app.resilientsystems.com/rest/orgs/201/incidents`
+        .. note::
+            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
+            if you specify a uri of ``/incidents``, the actual URL would be something like:
+            ``https://soar.ibm.com/rest/orgs/201/incidents``
 
         :param uri: Relative URI of the resource to post.
+        :type uri: str
         :param payload: A dictionary value to be posted.
-        :param co3_context_token: The Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :return: A dictionary or array with the value returned by the server.
+        :type payload: dict
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :param headers: Optional headers to include.
+        :type headers: dict
+        :return: A dictionary or list with the value returned by the server.
+        :rtype: dict | list
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         # Call post of BaseClient. Convert exception if there is any
         response = None
         try:
-            response = super(SimpleClient, self).post(uri, payload, co3_context_token, timeout)
+            response = super(SimpleClient, self).post(uri, payload, co3_context_token, timeout, headers=headers, skip_retry=skip_retry, **kwargs)
         except co3base.BasicHTTPException as ex:
             _raise_if_error(ex.get_response())
         return response
@@ -438,19 +565,26 @@ class SimpleClient(co3base.BaseClient):
         """
         PATCH request to the specified URI.
 
-        Note that this URI is relative to :samp:`<base_url>/rest/orgs/<org_id>`.  So for example,
-        if you specify a uri of :samp:`/incidents`, the actual URL would be something like:
-        `https://app.resilientsystems.com/rest/orgs/201/incidents`
+        .. note::
+            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
+            if you specify a uri of ``/incidents``, the actual URL would be something like:
+            ``https://soar.ibm.com/rest/orgs/201/incidents``
 
         :param uri: Relative URI of the resource to patch.
-        :param patch: The :class:`Patch` object to apply
-        :param co3_context_token: the Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :param overwrite_conflict: always overwrite fields in conflict.  Note that if True, the passed-in patch
-                object will be modified if necessary.
-        :return: The response object.
+        :type uri: str
+        :param patch: The :class:`~resilient.patch.Patch` object to apply
+        :type patch: :class:`~resilient.patch.Patch`
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :param overwrite_conflict: always overwrite fields in conflict. If ``True``, the passed-in patch
+            object will be modified if necessary.
+        :type overwrite_conflict: bool
+        :return: the ``response`` from the endpoint.
+        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
         :raises SimpleHTTPException: if an HTTP exception or patch conflict occurs.
-        :raises PatchStatusException: If the patch failed to apply (and overwrite_conflict is False).
+        :raises PatchConflictException: If the patch failed to apply (and overwrite_conflict is False).
         """
         if overwrite_conflict:
             # Re-issue patch with intent to overwrite conflicts.
@@ -467,13 +601,19 @@ class SimpleClient(co3base.BaseClient):
         the specified callback is invoked, allowing the caller to adjust the patch as necessary.
 
         :param uri: Relative URI of the resource to patch.
-        :param patch: The :class:`Patch` object to apply
+        :type str: Relative URI of the resource to patch.
+        :param patch: The :class:`~resilient.patch.Patch` object to apply
+        :type patch: :class:`~resilient.patch.Patch`
         :param callback: Function/lambda to invoke when a patch conflict is detected.  The function/lambda must be
-          of the following form: `def my_callback(response, patch_status, patch)`.
-          If your callback raises :class:`NoChange`, the update is skipped.
-        :param co3_context_token: the Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :return: The response object.
+          of the following form: ``def my_callback(response, patch_status, patch)``.
+          If your callback raises :class:`resilient.NoChange <resilient.co3base.NoChange>`, the update is skipped.
+        :type callback: func
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :return: the ``response`` from the endpoint.
+        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
         """
         response = self._patch(uri, patch, co3_context_token, timeout)
 
@@ -482,41 +622,67 @@ class SimpleClient(co3base.BaseClient):
 
         return response
 
-    def post_attachment(self, uri, filepath,
-                        filename=None, mimetype=None, data=None, co3_context_token=None, timeout=None):
+    def post_attachment(self, uri, filepath, filename=None,
+                        mimetype=None, data=None, co3_context_token=None, timeout=None,
+                        bytes_handle=None, skip_retry=[]):
         """
         Upload a file to the specified URI
-        e.g. "/incidents/<id>/attachments" (for incident attachments)
-        or,  "/tasks/<id>/attachments" (for task attachments)
+        e.g. ``/incidents/<id>/attachments`` (for incident attachments)
+        or ``/tasks/<id>/attachments`` (for task attachments)
+
+        .. warning::
+            Please see our updated :class:`resilient_lib.write_file_attachment <resilient_lib.components.resilient_common.write_file_attachment>`
+            as this method takes a data stream and is more reliable
 
         :param uri: Relative URI of the resource to post.
+        :type uri: str
         :param filepath: the path of the file to post
+        :type filepath: str
         :param filename: optional name of the file when posted
+        :type filename: str
         :param mimetype: optional override for the guessed MIME type
-        :param data: optional dict with additional MIME parts (not required for file attachments; used in artifacts)
-        :param co3_context_token: the Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
+        :type mimetype: str
+        :param data: optional dict with additional ``MIME`` parts (not required for file attachments; used in artifacts)
+        :type data: dict
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :param bytes_handle: BytesIO handle for content or use filepath
+        :type bytes_handle: BytesIO
+        :return: the ``response`` from the endpoint.
+        :rtype: `requests.Response <https://docs.python-requests.org/en/latest/api/#requests.Response>`_
         """
         # Call BaseClient post_attachment. Convert exception if there is any
         response = None
         try:
-            response = super(SimpleClient, self).post_attachment(uri, filepath, filename, mimetype, data,
-                                                                 co3_context_token, timeout)
+            response = super(SimpleClient, self).post_attachment(uri, filepath, filename=filename,
+                                                                 mimetype=mimetype,
+                                                                 data=data,
+                                                                 co3_context_token=co3_context_token,
+                                                                 timeout=timeout,
+                                                                 bytes_handle=bytes_handle,
+                                                                 skip_retry=skip_retry)
         except co3base.BasicHTTPException as ex:
             _raise_if_error(ex.get_response())
         return response
 
     def search(self, payload, co3_context_token=None, timeout=None):
         """
-        Posts to the SearchExREST endpoint.
+        Posts to the ``SearchExREST`` endpoint.
+
         Endpoint for performing full text searches through incidents and incident child objects
         (tasks, incident comments, task comments, milestones, artifacts, incident attachments,
         task attachments, and data tables).
 
-        :param payload: The SearchExInputDTO parameters for performing a search, as a dictionary
-        :param co3_context_token: the Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :return: List of results, as an array of SearchExResultDTO
+        :param payload: The SearchExInputDTO parameters for performing a search.
+        :type payload: dict
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :return: List of results, as an array of ``SearchExResultDTO``
+        :rtype: list
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         url = u"{0}/rest/search_ex".format(self.base_url)
@@ -533,20 +699,27 @@ class SimpleClient(co3base.BaseClient):
         return json.loads(response.text)
 
     def get_put(self, uri, apply_func, co3_context_token=None, timeout=None):
-        """Safely performs an update operation by a GET, calls your `apply_func` callback, then PUT
-        with the updated value.  If the put call returns a 409 error, these steps are retried.
+        """
+        Safely performs an update operation by a GET, calls your ``apply_func`` callback, then PUT
+        with the updated value.  If the put call returns a ``409`` error, these steps are retried.
 
-        Note that this URI is relative to :samp:`<base_url>/rest/orgs/<org_id>`.  So for example,
-        if you specify a uri of :samp:`/incidents`, the actual URL would be something like:
-        `https://app.resilientsystems.com/rest/orgs/201/incidents`
+        .. note::
+            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
+            if you specify a uri of ``/incidents``, the actual URL would be something like:
+            ``https://soar.ibm.com/rest/orgs/201/incidents``
 
         :param uri: Relative URI of the resource to get and update.
+        :type uri: str
         :param apply_func: A callback function that you implement to update the resource.  The function must be
-          of the following form: `def my_apply_func(object_to_update)`, and update the object.
-          If your callback raises :class:`NoChange`, the update is skipped.
-        :param co3_context_token: The Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :return: A dictionary or array with the value returned by the PUT operation.
+          of the following form: ``def my_apply_func(object_to_update)``, and update the object.
+          If your callback raises :class:`resilient.NoChange <resilient.co3base.NoChange>`, the update is skipped.
+        :type apply_func: func
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :return: A dictionary or list with the value returned by the PUT operation.
+        :rtype: dict | list
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         # Call BaseClient get_put. Convert exception if there is any
@@ -557,44 +730,59 @@ class SimpleClient(co3base.BaseClient):
             _raise_if_error(ex.get_response())
         return res
 
-    def put(self, uri, payload, co3_context_token=None, timeout=None):
-        """Directly performs an update operation by PUT to the specified URI.
+    def put(self, uri, payload, co3_context_token=None, timeout=None, headers=None, skip_retry=[]):
+        """
+        Directly performs an update operation by PUT to the specified URI.
+
+        .. note::
+            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
+            if you specify a uri of ``/incidents``, the actual URL would be something like:
+            ``https://soar.ibm.com/rest/orgs/201/incidents``
 
-        Note that this URI is relative to :samp:`<base_url>/rest/orgs/<org_id>`.  So for example,
-        if you specify a uri of :samp:`/incidents`, the actual URL would be something like:
-        `https://app.resilientsystems.com/rest/orgs/201/incidents`
 
         :param uri: Relative URI of the resource to update.
+        :type uri: str
         :param payload: The object to update.
-        :param co3_context_token: The Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
-        :return: A dictionary or array with the value returned by the server.
+        :type payload: dict
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
+        :param headers: Optional headers to include.
+        :type headers: dict
+        :return: A dictionary or list with the value returned by the server.
+        :rtype: dict | list
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         # Call BaseClient put. Convert exception if there is any
         response = None
         try:
-            response = super(SimpleClient, self).put(uri, payload, co3_context_token, timeout)
+            response = super(SimpleClient, self).put(uri, payload, co3_context_token, timeout, headers=headers, skip_retry=skip_retry)
         except co3base.BasicHTTPException as ex:
             _raise_if_error(ex.get_response())
         return response
 
-    def delete(self, uri, co3_context_token=None, timeout=None):
-        """Deletes the specified URI.
+    def delete(self, uri, co3_context_token=None, timeout=None, skip_retry=[]):
+        """
+        Deletes the specified URI.
 
-        Note that this URI is relative to :samp:`<base_url>/rest/orgs/<org_id>`.  So for example,
-        if you specify a uri of :samp:`/incidents`, the actual URL would be something like:
-        `https://app.resilientsystems.com/rest/orgs/201/incidents`
+        .. note::
+            This URI is relative to ``<base_url>/rest/orgs/<org_id>``.  So for example,
+            if you specify a uri of ``/incidents``, the actual URL would be something like:
+            ``https://soar.ibm.com/rest/orgs/201/incidents``
 
-        :param uri: Relative URI of the resource to update.
-        :param co3_context_token: The Co3ContextToken from an Action Module message, if available.
-        :param timeout: optional timeout (seconds)
+        :param uri: Relative URI of the resource to delete.
+        :type uri: str
+        :param co3_context_token: The ``Co3ContextToken`` from an Action Module message, if available.
+        :type co3_context_token: str
+        :param timeout: Optional timeout (seconds).
+        :type timeout: int
         :raises SimpleHTTPException: if an HTTP exception occurs.
         """
         # Call BaseClient delete. Convert exception if there is any
         response = None
         try:
-            response = super(SimpleClient, self).delete(uri, co3_context_token, timeout)
+            response = super(SimpleClient, self).delete(uri, co3_context_token, timeout, skip_retry=skip_retry)
         except co3base.BasicHTTPException as ex:
             _raise_if_error(ex.get_response())
         return response
